/*
 *  Biblioteka libekglog, odczyt i zarzdzanie baz wiadomoci i statusw.
 *  Copyright (C) 2004 Piotr Kalinowski <pitkali@o2.pl> All rights reserved.
 *  Niniejsza biblioteka podlega licencji GNU GPL.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <locale>
#include <sstream>
#include <fstream>
#include <utility>
#include <sys/time.h>
#include <unistd.h> // unlink...
#include <errno.h> // jak co unlinkowi nie pjdzie...
#include <cstring> // strerror
#include <cstdlib> // strtol

#include "ekglog.hpp"
#include "zfstream.h"
#include "directory.hpp"
using namespace EkgLog;
using namespace std;

// funkcja pomocnicza do sprawdzania poprawnoci indeksu, zwraca false w przypadku bdu
static inline bool
CheckIndex (string::size_type& idx, // indeks, ktry sprawdzamy
	    unsigned int line_no,   // numer linii logw, ktry jest analizowany
	    ostream& errors)	    // strumie gromadzcy informacje o bdach
{
  if (idx == string::npos) {
    errors << line_no << " ";
    return false;
    }
  return true;
}

// funkcja pomocnicza - okrela czy string naley zawrze w logach cudzysowach escapujc
// odpowiednie znaki. Jeli tak - zwraca true. Pobiera string i zakres do sprawdzenia
static inline bool
NeedToUnParse (const string& str, const string::size_type beg, const string::size_type end)
{
  register string::size_type i = beg;
  for (; i!=end; ++i)
    if ( (str[i] == '\'') || (str[i] == '"') || (str[i] == '\r') || (str[i] == '\n') || (str[i] == ',') )
      break;
  if (i != end)
    return true;
  return false;
}

// funkcja pomocnicza - Parsuje string eliminujc wyescapowane znaki i zastpujc je nimi samymi
static inline string
ParseString (const string& str, const string::size_type beg, const string::size_type end)
{
  /* beg - indeks pocztkowy, end - indeks za-kocowy
   parsujemy lini:
     \r (dwa znaki) -> \r (jeden znak)
     \n -> \n (jak wyej)
     \' -> '
     \" -> "
     \\ -> \  */
  string ret;
  for (register string::size_type i = beg,j = beg; j<end;) {
    while ((j<end) && (str[j] != '\\'))
      ++j;
    ret.append (str, i, j-i);
    if (j == end)
      break;
    if (j == end - 1) {
      ret.append (1, '\\');
      break;
    }
    switch (str[j+1]) {
      case 'r': // omi \r
	break;
      case 'n': // nowa linia
	ret.append (1, '\n'); break;
      case '\'': // hmm - on escapuje apostrofy?
	ret.append (1, '\''); break;
      case '"':
	ret.append (1, '"'); break;
      case '\\':
	ret.append (1, '\\'); break;
      default:
	ret.append (1, '\\');
	ret.append (1, str[j+1]);
	break;
    }
    j += 2;
    i = j;
  }
  return ret;
}

// funkcja pomocnicza - Odwrotno funkcj ParseString
// niezmiennik: UnParseString(ParseString(str)) == str (pominito indeksy dla czytelnoci)
static inline string
UnParseString (const string& str, const string::size_type beg, const string::size_type end)
{
  // odparsowujemy lini na uytek zapisu logw. Dla spjnoci interfejsu sygnatura zawiera
  // pocztek i koniec (indeks pocztkowy i za-kocowy), cho obecnie funkcja save nie
  // skorzysta na tym (waciwie to writeMsg i writeStatus).
  //
  // Escapujemy znaki ', ", \, \n (ten ostatni zamieniamy zgodnie z konwencj ekg na \r\n)
  
  string ret;
  for (register string::size_type i = beg,j = beg; j<end;) {
    while ( (j<end) && (str[j] != '\'') && (str[j] != '"') && (str[j] != '\\') && (str[j] != '\n') )
      ++j;
    ret.append (str, i, j-i);
    if (j == end)
      break;
    ret.append (1, '\\'); // escape character
    if (str[j] == '\n')
      ret.append ("r\\n");
    else
      ret.append (1, str[j]);
    i = ++j;
  }
  return ret;
}
    
// funkcja pomocnicza - wczytuje plik
static inline void
LoadFile (map<const uin_t, UinLog>& uins, // mapa uinw
	  const string& file,		  // plik do wczytania
	  string& errors_str,		  // tu zostan zapisane bdy zwizane z tym plikiem
	  unsigned int& msg_no, unsigned int& stat_no, // referencje do statystyk odczytu
	  const dBase::log_format_t format = dBase::FILE, // format bazy
	  const uin_t uin = 0) // jeli format katalogowy - jaki wczytujemy uin? (bo nie czytamy go
			      // wtedy z pliku.
  throw(runtime_error)
{
  gzifstream input (file.c_str());
  if (input.fail())
    throw runtime_error ("Nie udao si otworzy pliku z logiem!");
  char buf[8192]; // bufor na lini
  unsigned int line_no = 0; // do informacji o bdach
  Message msg;
  Status stat;

  // wykorzystywane do przetwarzania wartoci numerycznych.
  char *endptr;
  uin_t cur_uin;
  log_iterator uin_iter;
  pair<log_iterator, log_iterator> search_res;
  time_t time;

  ostringstream errors;

  if ((format == dBase::DIR) || (format == dBase::DIRGZ)) {
    // dla formatu katalogowego pobierz za wczasu iterator na UinLog, do ktrego
    // wstawiamy wiadomoci i statusy.
    cur_uin = uin;
    uin_iter = uins.insert (make_pair(cur_uin, UinLog(cur_uin))).first;
  }
  
  while (input.peek() != gzifstream::traits_type::eof()) {
    // wczytujemy lini do bufora - to chyba najszybszy sposb ze strumieniami
    input.getline (buf, 8192);
    string line (buf);
    while (!input.eof() && input.fail()) {
      // jest co jeszcze do danej linii
      input.clear();
      input.getline (buf, 8192);
      line += buf;
    }

    ++line_no;

    // --- linia jest - teraz PRZETWARZANIE ---
    // odczytaj typ wpisu
    string::size_type idx = line.find_first_of (",");
    if ( (line[0] == 'm') || (line[0] == 'c')) {
      // --- wiadomo ---
      if (!CheckIndex(idx, line_no, errors))
	continue;
      string tmp (line, idx-4, 4);
      if (tmp == "send")
	msg.type (MSG_SENT);
      else
	msg.type (MSG_RECV);
      // odczytaj uin
      string::size_type idx2 = line.find_first_of (",", ++idx);
      if (!CheckIndex(idx2, line_no, errors))
	continue;
      if ((format == dBase::FILE) || (format == dBase::FILEGZ)) {
	cur_uin = strtol (&line[idx], &endptr, 10);
	if (endptr == &line[idx]) {
	  errors << line_no << " ";
	  continue;
	}
	// czy mamy taki uin w logach?
	search_res = uins.equal_range(cur_uin);
	if (search_res.first == search_res.second) // ups, nie ma jeszcze takiego
	  uin_iter = uins.insert (make_pair(cur_uin, UinLog(cur_uin))).first;
	else
	  uin_iter = search_res.first;
      }
      idx = idx2 + 1;
      // pomi nick (uywamy mapowania podanego przez klienta biblioteki)
      idx = line.find_first_of (",", idx);
      if (!CheckIndex(idx, line_no, errors))
	continue;
      // wczytaj pierwszy czas
      idx2 = line.find_first_of (",", ++idx);
      if (!CheckIndex(idx2, line_no, errors))
	continue;
      time = strtol(&line[idx], &endptr, 10);
      if (endptr == &line[idx]) {
	errors << line_no << " ";
	continue;
      }
      if (msg.type() == MSG_SENT)
	msg.sent (time);
      // wczytaj czas wysania, jeli jest to wiadomo odebrana
      else if (msg.type() == MSG_RECV) {
	msg.recv (time);
	idx = idx2 + 1;
	idx2 = line.find_first_of (",", idx);
	if (!CheckIndex(idx2, line_no, errors))
	  continue;
	time = strtol (&line[idx], &endptr, 10);
	if (endptr == &line[idx]) {
	  errors << line_no << " ";
	  continue;
	}
	msg.sent (time);
      }
      // idx2 = ostatni przecinek. po nim ju jest tylko wiadomo.
      // 
      // UWAGA: pono w bardzo starych logach, wiadomoci wielolinijkowe
      // byy zapisywane w wielu liniach - rzeczywicie kiedy si na to
      // natknem. Tym niemniej NIE WSPIERAMY takiego formatu. Przed
      // odczytem tak starej bazy trzeba j przekonwertowa.

      ++idx2; // pierwszy znak wiadomoci

      // Docz biec lini do treci wiadomoci
      if (line[idx2] == '"')
	// pomi cudzysowy
	msg.msg (ParseString(line, idx2+1, line.size()-1));
      else
	// nie parsuj jak nie ma cudzyslowia - nie ma czego
	msg.msg (line, idx2, line.size());

      uin_iter->second.addMessage (msg);
      ++msg_no;
    } // if wiadomo
    else if (line.compare(0,2,"st") == 0) {
      // --- status ---
      // odczytaj uin
      string::size_type idx2 = line.find_first_of (",", ++idx);
      if (!CheckIndex(idx2, line_no, errors))
	continue;
      if ((format == dBase::FILE) || (format == dBase::FILEGZ)) {
	// pobierz numerek uin wpisu
	cur_uin = strtol(&line[idx], &endptr, 10);
	if (endptr == &line[idx]) {
	  errors << line_no << " ";
	  continue;
	}
	// czy mamy taki uin w logach?
	search_res = uins.equal_range(cur_uin);
	if (search_res.first == search_res.second) // ups, nie ma jeszcze takiego
	  uin_iter = uins.insert (make_pair(cur_uin, UinLog(cur_uin))).first;
	else
	  uin_iter = search_res.first;
      }
      idx = idx2 + 1;
      // pomi nick
      idx = line.find_first_of (",", idx);
      if (!CheckIndex(idx, line_no, errors))
	continue;
      // pomi IP
      idx = line.find_first_of (",", ++idx);
      if (!CheckIndex(idx, line_no, errors))
	continue;
      // odczytaj czas
      idx2 = line.find_first_of (",", ++idx);
      if (!CheckIndex(idx2, line_no, errors))
	continue;
      time = strtol (&line[idx], &endptr, 10);
      if (endptr == &line[idx]) {
	errors << line_no << " ";
	continue;
      }
      stat.time (time);
      // typ statusu
      idx = idx2+1;
      idx2 = line.find_first_of (",", idx);
      string stype;
      if (idx2 == string::npos)
	stype.assign (line, idx, line.size()-idx);
      else
	stype.assign (line, idx, idx2 - idx);
      
      if (stype == "notavail")
	stat.type (NOTAVAILABLE);
      else if (stype == "busy")
	stat.type (BUSY);
      else if (stype == "avail")
	stat.type (AVAILABLE);
      else
	stat.type (INVISIBLE);

      if (idx2 != string::npos) {
	// opis statusu.
	idx = idx2+1;
	if (line[idx] == '"')
	  // pomi cudzysowy
	  stat.desc (ParseString(line, idx+1, line.size()-1));
	else
	  // nie parsuj jak nie ma cudzyslowa - nie ma czego
	  stat.desc (line, idx, line.size());
      }
      else
	stat.desc ("");
      uin_iter->second.addStatus (stat);
      ++stat_no;
    } // if status

  } // while jest co czyta
  input.close();

  if (errors.str().size() != 0) {
    // errors reported
    errors_str += "\nPlik: ";
    errors_str += file;
    errors_str += ", linie: ";
    errors_str += errors.str();
  }

} // LoadFile

// otwrz logi - dokumentacja w ekglog.hpp
bool
dBase::open(const string& _path, const dBase::log_format_t format,
    const time_t interval) throw(runtime_error)
{
  // jeli jakie logi ju s, open je nadpisuje
  if (!_logs.empty()) {
    _logs.clear();
    _messages = _statuses = 0;
    _time = 0;
    _badbit = true;
  }
  
  // Bdziemy zbiera bdy formatu - inicjuj wiadomo
  _error_msg = "";
  bool retval = true;

  string path = _path; // lokalna kopia argumenty - bdziemy ni potencjalnie onglowa,
                       // a argument ma kwalifikator const (waciwie mona by pomyle
		       // nad przekazywaniem go przez warto, skoro i tak trzeba go 
		       // skopiowa...
  userlist_t::const_iterator nickname;
  
  // Mierzymy czas
  struct timeval start, finish;
  gettimeofday (&start, NULL);
  
  if ((format == DIR) || (format == DIRGZ)) {
    Directory dir (path);
    if (dir.isbad())
      throw runtime_error (dir.get_error().c_str());
    
    string filename = dir.read_entry();
    while (filename.size()) {
      uin_t uin;
      istringstream i (filename);
      i >> uin;
      if (!i.fail()) {
	LoadFile (_logs, path + "/" + filename, _error_msg, _messages, _statuses, format, uin);
      }
      filename = dir.read_entry();
    }
    dir.close();
  }
  else {
    if (format == FILEGZ) {
      path.append (".gz"); // w konfigu ekg path powinien by podany bez rozszerzenia.
			   // biblioteka trzyma si tej konwencji.
    }
    LoadFile (_logs, path, _error_msg, _messages, _statuses);
  }
  // ustaw listy czatw i nickw.
  for (log_iterator i = _logs.begin(); i!=_logs.end(); ++i) {
    i->second.fillChats(interval);
    nickname = _nicks.find (i->second.uin());
    if (nickname != _nicks.end())
      i->second.nick(nickname->second);
  }
  _badbit = false; // obiekt dBase jest ju poprawny

  gettimeofday (&finish, NULL);
  _time = finish.tv_sec-start.tv_sec + 0.000001*(finish.tv_usec-start.tv_usec);

  if (_error_msg.size() != 0) {
    // wystpiy bdy formatu
    _error_msg = "Wystpiy bdy formatu logw!" + _error_msg;
    retval = false;
  }
  return retval;
}

// funkcja pomocnicza - zapisz wiadomo do strumienia
static inline void 
writeMsg (ostream& out, const Message& msg, const uin_t uin, const string& nick) 
{
  msg_type t = msg.type();
  if (t == MSG_RECV)
    out << "chatrecv,";
  else
    out << "chatsend,";
  out << uin << "," << nick << ",";
  if (t == MSG_RECV)
    out << msg.recv() << ",";
  out << msg.sent() << ",";
  if (NeedToUnParse(msg.msg(), 0, msg.msg().size()))
    out << '"' << UnParseString(msg.msg(), 0, msg.msg().size()) << '"';
  else
    out << msg.msg();
  out << "\n";
}

// jak wyej, ale dla statusu
static inline void
writeStatus (ostream& out, const Status& status, const uin_t uin, const string& nick)
{
  out << "status," << uin << "," << nick << ",0.0.0.0:0,";
  out << status.time() << ",";
  status_type t = status.type();
  switch (t) {
    case AVAILABLE:
      out << "avail,"; break;
    case BUSY:
      out << "busy,"; break;
    case INVISIBLE:
      out << "invisible,"; break; // UWAGA: waciwie nigdy tego nie sprawdziem, z braku
				  // z braku moliwoci. Moe to w sumie nie jest takie wane
				  // bo i tak prawdopodobiestwo wystpienia takiego wpisu
				  // jest znikome ze wzgldu na protok, a podczas odczytu
				  // przez niniejsz baz zostanie i tak poprawne rozpoznane.
    case NOTAVAILABLE:
      out << "notavail,"; break; 
  }
  if (NeedToUnParse(status.desc(), 0, status.desc().size()))
    out << '"' << UnParseString(status.desc(), 0, status.desc().size()) << '"';
  else
    out << status.desc();
  out << "\n";
}

// kryterium sortowania dla listy wiadomoci do zapisania (chronologiczne)
bool lt_logline (const pair<time_t, string>& l, const pair<time_t, string>& r) {
  return (l.first < r.first);
}

// zapisz baz do pliku - dokumentacja w nagwku ekglog.hpp
bool
dBase::save (const string& path, const log_format_t format) throw(runtime_error)
{
  if (_badbit)   // nie ma czego zapisywa, wic...
    return true; // ...waciwie operacja si udaa
  
  bool retval = true;
  
  if ((format == DIR) || (format == DIRGZ)) {
    // kady uin w osobnym pliku - path jest katalogiem.
    for (const_log_iterator i = _logs.begin(); i!=_logs.end(); ++i) {
      ostringstream filename;
      filename << path << '/' << i->second.uin();
      ostream* file;
      try {
	if (format == DIRGZ) {
	  filename << ".gz";
	  file = new gzofstream (filename.str().c_str(), ios::out|ios::trunc);
	}
	else
	  file = new ofstream (filename.str().c_str(), ios::out|ios::trunc);
      }
      catch (bad_alloc) {
	throw runtime_error ("Brak pamici operacyjnej na strumie plikowy!\n");
      }
      if ( *file == false) {
	ostringstream error;
	error << "Nie mona otworzy pliku: " << filename.str() << endl;
	throw runtime_error (error.str());
      }
      const_msg_iterator msg_it = i->second.msgBegin();
      const_status_iterator stat_it = i->second.statusBegin();
      while ((msg_it != i->second.msgEnd()) && (stat_it != i->second.statusEnd())) {
	if (msg_it->recv() < stat_it->time()) {
	  writeMsg(*file, *msg_it, i->second.uin(), i->second.nick());
	  ++msg_it;
	}
	else {
	  writeStatus(*file, *stat_it, i->second.uin(), i->second.nick());
	  ++stat_it;
	}
      }
      // co zostao - trzeba to przepisa do koca. tylko jedna z poniszych ptli
      // wykona si cho raz.
      while (msg_it != i->second.msgEnd()) {
	writeMsg(*file, *msg_it, i->second.uin(), i->second.nick());
	++msg_it;
      }
      while (stat_it != i->second.statusEnd()) {
	writeStatus(*file, *stat_it, i->second.uin(), i->second.nick());
	++stat_it;
      }
      // cay uin zapisany.
      delete file;
    } // for each uin
    // Teraz sprawdzamy, czy nie trzeba jaki plikw usun (z powodu usunicia wszystkich
    // wpisw, ktre miayby tam si znale)
    //
    // Algorytm: dla kadego logu w plikach sprawdzamy, czy jest obecny w bazie. (brute)
    Directory dir(path);
    if (dir.isbad()) {
      // mao prawdopodobne, ale na wszelki wypadek...
      // katalogu nie otworzono, wic nie mona nic zrobi - zapisz bd w zmiennej _error_msg
      _error_msg = "Bd przy weryfikacji zawartoci katalogu historii wiadomoci:\n";
      _error_msg += dir.get_error();
    }
    else {
      string filename = dir.read_entry();
      while (filename.size()) {
	// pobierz uin
	istringstream in (filename);
	uin_t uin;
	in >> uin;
	if (!in.fail() && (log(uin) == end())) {
	  string remove = path + "/" + filename;
	  if (format == DIRGZ)
	    remove += ".gz";
	  if (unlink(remove.c_str()) < 0 )
	    _error_msg += strerror(errno);
	}
	filename = dir.read_entry();
      }
    }
    dir.close();
  }
  else {
    // format == FILE || format == FILEGZ
    // Algorytm:
    //   wrzucamy wszystko do jednej listy, sortujc po czasach (osobno dla statusw
    //   i wiadomoci). Potem postpujemy jak wyej.
    
    // otwrz plik
    ostream* file;
    try {
      if (format == FILEGZ)
	file = new gzofstream( string(path + ".gz").c_str(), ios::out|ios::trunc );
      else
	file = new ofstream(path.c_str(), ios::out|ios::trunc);
    }
    catch (bad_alloc) {
      throw runtime_error ("Brak pamici operacyjnej na strumie plikowy!\n");
    }
    if ( *file == false ) {
      ostringstream error;
      error << "Nie mona otworzy pliku: " << path;
      if (format == FILEGZ)
	error << ".gz";
      error << endl;
      throw runtime_error (error.str());
    }
    
    list<pair<time_t, string> > messages, statuses;

    for (const_log_iterator i = _logs.begin(); i!=_logs.end(); ++i) {
      ostringstream log_line;
      string empty_string = "";
      for (const_msg_iterator j = i->second.msgBegin(); j!=i->second.msgEnd(); ++j) {
	writeMsg(log_line, *j, i->second.uin(), i->second.nick());
	messages.push_back( make_pair(j->recv(), log_line.str()) );
	log_line.str(empty_string);
      }
      for (const_status_iterator j = i->second.statusBegin(); j!=i->second.statusEnd(); ++j) {
	writeStatus(log_line, *j, i->second.uin(), i->second.nick());
	statuses.push_back( make_pair(j->time(), log_line.str()) );
	log_line.str(empty_string);
      }
    }

    // sortuj listy
    messages.sort(lt_logline);
    statuses.sort(lt_logline);

    // zapisz w pliku
    list<pair<time_t, string> >::iterator msg_it = messages.begin();
    list<pair<time_t, string> >::iterator stat_it = statuses.begin();
    while ((msg_it != messages.end()) && (stat_it != statuses.end())) {
      if (msg_it->first < stat_it->first) {
	(*file) << msg_it->second;
	++msg_it;
      }
      else {
	(*file) << stat_it->second;
	++stat_it;
      }
    }
    // co zostao - trzeba to przepisa do koca. tylko jedna z poniszych ptli
    // wykona si cho raz.
    while (msg_it != messages.end()) {
      (*file) << msg_it->second;
      ++msg_it;
    }
    while (stat_it != statuses.end()) {
      (*file) << stat_it->second;
      ++stat_it;
    }
    // cay uin zapisany.
    delete file;

      
  } // format == FILE(GZ)

  return retval;
      
} // dBase::save

// operacja scalania - patrz ekglog.hpp
bool 
  dBase::merge (const string& path, const log_format_t format, const time_t interval)
throw(runtime_error)
{
  if (_badbit) // nie ma czym scala, wic po prostu otwrz histori wiadomoci...
    return open(path, format, interval);

  struct timeval start, finish;
  gettimeofday(&start, NULL);

  // ETAP 1 - wczytaj nowe logi
  dBase to_merge;
  bool retval;
  try {
    retval = to_merge.open (path, format, _nicks, interval);
  } catch (runtime_error& err) {
    string error_msg = "Bd podczas otwierania historii wiadomoci do scalenia: ";
    error_msg += err.what();
    throw runtime_error (error_msg.c_str());
  }

  // Ustaw informacj o bdach formatu w logach do scalenia.
  if (!retval)
    _error_msg = to_merge.get_error();

  // ETAP 2 - waciwe scalanie
  for(const_log_iterator log_to_merge = to_merge.begin();
      log_to_merge != to_merge.end();
      ++log_to_merge) // dla logu kadego uin do scalenia
  {
    log_iterator dest_log = log (log_to_merge->first); // znajd docelowy dla tego uina
    if (dest_log == end()) {
      // cae logi tego uina s nowe - po prostu trzeba je doda
      pair<log_iterator, bool> pos = _logs.insert (*log_to_merge);
      pos.first->second.fillChats(interval);
      _messages += pos.first->second.msgCount();
      _statuses += pos.first->second.statusCount();
    } else {
      // ETAP 2A - scalanie wiadomoci
      
      // tylko cz wiadomoci (jeli jakie) naley doda
      msg_iterator compare_with = dest_log->second.msgBegin();
      for(const_msg_iterator msg_to_merge = log_to_merge->second.msgBegin();
	  msg_to_merge != log_to_merge->second.msgEnd();
	  ++msg_to_merge) // dla kadej wiadomoci do scalenia...
      {
	// znajd wiadomo o czasie odebrania niepniejszym (ostatnia odebrana przed lub o takim samym)
	// UWAGA: pierwszym kluczem uporzdkowania w logach jest czas odebrenia wiadomoci. Dopiero potem
	// jest czas wysania (nie jest prawd tak zupenie i do koca, ale poniewa dla wiadomoci wysanych
	// te czasy s takie same dla bazy, jest to wystarczajce do uporzdkowania wiadomoci).
	for (; 
	    (compare_with != dest_log->second.msgEnd()) && (compare_with->recv() < msg_to_merge->recv());
	    ++compare_with)
	  ;
	if (compare_with == dest_log->second.msgEnd()) {
	  // nie znaleziono wiadomoci niepniejszych, tylko wczeniejsze
	  // wniosek: wszystkie wiadomoci do scalenia s nowsze i naley
	  //          je doda
	  dest_log->second.addMessage(msg_to_merge, log_to_merge->second.msgEnd());
	  ++_messages;
	  break;
	} else if (compare_with->recv() > msg_to_merge->recv()) {
	  // znalelimy wiadomo pniejsz, ergo bieca wiadomo nie
	  // wystpuje w naszej historii wiadomoci
	  // wniosek: doda j
	  dest_log->second.addMessage(compare_with, *msg_to_merge);
	  ++_messages;
	} else {
	  // Znalelimy wiadomo odebran w tym samym czasie. Teraz sprawdzamy,
	  // czy jaka wiadomo tego typu jest identyczna z wiadomoci do dodania.
	  // w przeciwnym razie dodajemy j.
	  for (;
	      (compare_with != dest_log->second.msgEnd()) // dopki w logach
	      && (compare_with->recv() == msg_to_merge->recv()); // dopki odebrano w tym samym czasie
	      ++compare_with)
	  {
	    if (   (compare_with->type() == msg_to_merge->type())
		&& (compare_with->sent() == msg_to_merge->sent())
		&& (compare_with->msg() == msg_to_merge->msg())   )
	      // wiadomoci s takie same - koniec przeszukiwania, ju tak mamy
	      break;
	  }
	  if (compare_with == dest_log->second.msgEnd()) {
	    // poza logami, zatem nie ma tam naszej wiadomoci, doda j
	    dest_log->second.addMessage(*msg_to_merge);
	    ++_messages;
	  }
	  else if (   (compare_with->recv() != msg_to_merge->recv()) ) {
	    // ominlimy wiadomoci o takim samym czasie odebrania i nie ma naszej
	    // wniosek: doda j
	    
	    // wiadomo compare_with jest odebrana pniej (wczeniejsze i te same ju ominlimy).
	    // Gdyby to bya pierwsza wiadomo w logu, zostaoby to wyapane wczeniej i bymy nie
	    // byli w tym miejscu - wniosek: compare_with NIE JEST pierwsz wiadomo i mona bezpiecznie
	    // co najmniej raz zmniejszy ten iterator.
	    
	    // ustalam miejsce wstawienia, aby si zgadzay czasy wysania
	    for(--compare_with; // wejd w obszar wiadomoci o takim samym czasie odebrania
		(compare_with != dest_log->second.msgBegin()) // nie chcemy cofn si przed pocztek logw...
		&& (compare_with->recv() == msg_to_merge->recv()) // pozosta wrd odebranych razem z nasz
		&& (compare_with->sent() > msg_to_merge->recv()); // omi wysane pniej od naszej
		--compare_with)
	      ;
	    if (compare_with == dest_log->second.msgBegin()) {
	      // mamy pierwsz wiadomo. nie zostaa ona sprawdzona, musimy to zrobi teraz, aby ustali,
	      // czy naley nasz wiadomo doda pzed czy za ni.
	      if (   (compare_with->recv() >= msg_to_merge->recv())   // odebrana niewczeniej
		  && (compare_with->sent() > msg_to_merge->sent())  ) { // wysana pniej
		dest_log->second.addMessage(compare_with, *msg_to_merge); // wstaw przed ni
		++_messages;
	      }
	    }
	    
	    // compare_with wskazuje wiadomo odebran wczeniej. naley go zwikszy o jeden aby wej
	    // do obszaru odebranych w tym samy czasie co nasza wiadomo
	    // LUB
	    // compare_with wskazuje wiadomo wysan wczeniej lub w tym samym czasie - zwikszamy
	    // compare_with, aby doda j na samym kocu wiadomoci wysanych o tym samym czasie lub
	    // wstawi przed pierwsz wiadomo o pniejszym czasie (za ostatani o czasie wczeiejszym)
	    ++compare_with;
	    dest_log->second.addMessage(compare_with, *msg_to_merge);
	    ++_messages;
	  }

	  // jeeli aden z powyszych warunkw nie zosta speniony, tzn, e ptla
	  // zostaa przerwana przez instrukcj break wewntrz, czyli mamy ju tak
	  // wiadomo i nie potrzebujemy jej dodawa.
	  
	} // wiadomoci w tym samym czasie
      } // wiadomoci do scalenia

      // ETAP 2B - scalania statusw.
      status_iterator st_compare_with = dest_log->second.statusBegin();
      for(const_status_iterator status_to_merge = log_to_merge->second.statusBegin();
	  status_to_merge != log_to_merge->second.statusEnd();
	  ++status_to_merge)
      {
	// znajd statusy o takim samym czasie (lub pniejszym)
	for(;
	    (st_compare_with != dest_log->second.statusEnd())
	    && (st_compare_with->time() < status_to_merge->time());
	    ++st_compare_with)
	  ;
	if (st_compare_with == dest_log->second.statusEnd()) {
	  // ominlimy wszystkie statusy i nic - naley doda wszytkie pozostae
	  dest_log->second.addStatus(status_to_merge, log_to_merge->second.statusEnd());
	  ++_statuses;
	  break;
	}
	else if (st_compare_with->time() > status_to_merge->time()) {
	  // znalelimy status pniejszy, nie ma o takim samym czasie, wic mona go doda
	  dest_log->second.addStatus(st_compare_with, *status_to_merge);
	  ++_statuses;
	}
	else {
	  // status o takim samym czasie - sprawdmy, czy ktry nie jest taki sam jak nasz
	  for(;
	      (st_compare_with != dest_log->second.statusEnd())
	      && (st_compare_with->time() == status_to_merge->time());
	      ++st_compare_with)
	  {
	    if (   (st_compare_with->type() == status_to_merge->type())
		&& (st_compare_with->desc() == status_to_merge->desc())
	       )
	      break;
	  }
	  if (st_compare_with == dest_log->second.statusEnd()) {
	    // koniec, a naszego nie ma, ergo doda go
	    dest_log->second.addStatus(*status_to_merge);
	    ++_statuses;
	  }
	  else if (st_compare_with->time() != status_to_merge->time()) {
	    // nie ma naszego
	    dest_log->second.addStatus(st_compare_with, *status_to_merge);
	    ++_statuses;
	  }
	  
	  // ptl przerwao znalezienie takiego samego. Nie dodajemy go zatem

	} // statusy o takim samym czasie
	  
      } // scalono statusy
      dest_log->second.fillChats(interval);
    } // sprawdzono uin
  } // sprawdzono wszystkie uiny

  gettimeofday (&finish, NULL);
  _time += finish.tv_sec - start.tv_sec +
	    0.000001*(finish.tv_usec - start.tv_usec);

  return retval;
  
} // dBase::merge

